////////////////////////////////////////////////////////////////////////
// Class:       MichelEfinder
// Module Type: producer
// File:        MichelEfinder_module.cc
//
// Generated at Sun Jun 19 05:00:16 2016 by Andrey Sheshukov using artmod
// from cetpkgsupport v1_10_02.
// 
// Michel electron finding by Matt Strait, separated in a module
// For tests
//
////////////////////////////////////////////////////////////////////////

#include "art/Framework/Core/EDProducer.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Principal/Handle.h"
#include "art/Framework/Principal/Run.h"
#include "art/Framework/Principal/SubRun.h"
#include "art/Utilities/InputTag.h"
#include "fhiclcpp/ParameterSet.h"
#include "messagefacility/MessageLogger/MessageLogger.h"

#include "DDTBaseDataProducts/BaseProducts.h"
#include "DDTBaseDataProducts/HitList.h"
#include "DDTBaseDataProducts/DAQHit.h"
#include "DDTBaseDataProducts/HitSet.h"

#include "DDTBaseDataProducts/CompareDAQHit.h"

#include <memory>
#include <algorithm>

namespace novaddt {
  class MichelEfinder;
}

class novaddt::MichelEfinder : public art::EDProducer {
public:
  explicit MichelEfinder(fhicl::ParameterSet const & p);
  // The destructor generated by the compiler is fine for classes
  // without bare pointers or other resource use.

  // Plugins should not be copied or assigned.
  MichelEfinder(MichelEfinder const &) = delete;
  MichelEfinder(MichelEfinder &&) = delete;
  MichelEfinder & operator = (MichelEfinder const &) = delete;
  MichelEfinder & operator = (MichelEfinder &&) = delete;

  // Required functions.
  void produce(art::Event & e) override;
private:
    bool michelclose(const novaddt::DAQHit &trackend,
                     const novaddt::DAQHit &testcell);
    void findmichelexclusion( const novaddt::HitList & ends,
                              const novaddt::HitList & allhits,
                                    novaddt::HitList & answer);


private:
    const int kTDC_PER_USEC = 64;

    art::InputTag fTracksTag;
    art::InputTag fHitsTag;

    double fMaxDistance;
    int fDTime_min;
    int fDTime_max;
};


// Returns -1 if the testcell is later than the trackend by more than the
// time window.  Otherwise returns 0 if the testcell is not within the 
// michel exclusion zone and 1 if it is.


bool novaddt::MichelEfinder::michelclose(const novaddt::DAQHit & __restrict__ trackend,
                       const novaddt::DAQHit & __restrict__ testcell)
{
    if(testcell.View()!=trackend.View())return false;
 
     const double dplane = 1.5*(trackend.Plane().val - testcell.Plane().val);
     if(dplane > fMaxDistance) return false;

     const double dcell = trackend.Cell().val - testcell.Cell().val;
     if(dcell > fMaxDistance) return false;

     double dist2 = pow(dplane, 2) + pow(dcell, 2);

    // Exclude a disc of hits up to N cells = N/1.5 planes away.
    return dist2 <= fMaxDistance*fMaxDistance;

  // Now look, this is a very blunt cut, but we could get clever and notice
  // that Michel hits ought to be in a line connected to the track end and
  // < 5mus, whereas neutron hits are more random and can be much later, and
  // B-12 hits, if we care, are 1-2 cells again connected to the track end.
  // But let's see if this is sufficient before getting fancy.
}


//a comparison function for searching in time
bool cmp_tdc(const novaddt::DAQHit& h, const novaddt::TDC& tdc){
    return h.TDC()<tdc;
}





void novaddt::MichelEfinder::findmichelexclusion(
            const novaddt::HitList & ends,
            const novaddt::HitList & allhits,
                  novaddt::HitList & answer
)
{
  // Didn't find any track ends, nothing to exclude
  if(ends.empty()) return;
  
  for(const auto & end: ends){
      const int endview = end.View().val;

    //find the time range
    // Only need to look at the time region around the end
    auto tmin=end.TDC()+fDTime_min;
    auto tmax=end.TDC()+fDTime_max;
    const auto range_beg = std::lower_bound(allhits.begin(),allhits.end(),tmin, cmp_tdc);
    const auto range_end = std::lower_bound(range_beg,      allhits.end(),tmax, cmp_tdc);
    for(auto hit = range_beg; hit != range_end; hit++){
        bool res = michelclose(end, *hit);
        if(res) answer.push_back(*hit);
      }
    }
}

// 2D track: If in the Y view, assume downwards-going.  Find the lowest, and
// therefore last, hit.  If in the X view, can't tell which way the track is
// going.  Return both ends.
static std::vector<novaddt::DAQHit>
findtrackends2d(const std::vector<novaddt::DAQHit> & trackhits)
{
  std::vector<novaddt::DAQHit> answer;

  // Just in case, bail on no track hits
  if(trackhits.empty()) return answer;

  // This is a track in the Y view
  if(trackhits[0].View().val == daqchannelmap::Y_VIEW){
    // According to doc-11570, the bottom y cell has number zero.
    novaddt::DAQHit * lastycell = NULL;
    {
      int lowest = 1000; 
      for(const auto & hit: trackhits){
        if(hit.Cell().val < lowest){
          lowest = hit.Cell().val;
          lastycell = (novaddt::DAQHit*)&hit;
        }
      }
    }

    answer.push_back(*lastycell);
    return answer;
  }

  // This is a track in the X view.  Find each end.
  for(int increasingplane = 0; increasingplane < 2; increasingplane++){
    novaddt::DAQHit * lastxcell = NULL;
    {
      int endiest = increasingplane? 0: 10000;
      for(const auto & hit: trackhits){
        if(increasingplane ^ (hit.Plane().val < endiest)){
          endiest = hit.Plane().val;
          lastxcell = (novaddt::DAQHit*)&hit;
        }
      }
    }
    answer.push_back(*lastxcell);
  }

  return answer;
}


// 3D track: Assume track is downwards-going. Find lowest, and therefore last,
// Y-hit and from the Z-direction of the track, find the corresponding last
// X-hit
__attribute__((unused)) static std::vector<novaddt::DAQHit>
findtrackends3d(const std::vector<novaddt::DAQHit> & trackhits)
{
  std::vector<novaddt::DAQHit> answer;

  // Just in case, bail on no track hits
  if(trackhits.empty()) return answer;

  // According to doc-11570, the bottom y cell has number zero.
  novaddt::DAQHit * lastycell = NULL;
  {
    int lowest = 1000; 
    for(const auto & hit: trackhits){
      if(hit.View().val != daqchannelmap::Y_VIEW) continue;
      if(hit.Cell().val < lowest){
        lowest = hit.Cell().val;
        lastycell = (novaddt::DAQHit*)&hit;
      }
    }
  }

  // No y hits, return empty vector
  if(lastycell == NULL) return answer;
  answer.push_back(*lastycell);

  // Find out which way the track is going
  novaddt::DAQHit * someotheryhit = NULL;
  for(const auto & hit: trackhits){
    if(hit.View().val != daqchannelmap::Y_VIEW) continue;
    if(lastycell != (novaddt::DAQHit*)&hit)
      someotheryhit = (novaddt::DAQHit*)&hit;
  }

  const bool increasingplane = someotheryhit == NULL || 
    lastycell->Plane().val > someotheryhit->Plane().val;

  // Find ends in x.
  novaddt::DAQHit * lastxcell = NULL;
  {
    int endinest = increasingplane? 0: 10000;
    for(const auto & hit: trackhits){
      if(hit.View().val != daqchannelmap::X_VIEW) continue;

      if(increasingplane ^ (hit.Plane().val < endinest)){
        endinest= hit.Plane().val;
        lastxcell = (novaddt::DAQHit*)&hit;
      }
    }
  }

  // y hits, but no x hits, return vector of size one.
  if(lastxcell == NULL) return answer;
  answer.push_back(*lastxcell);

  return answer;
}



novaddt::MichelEfinder::MichelEfinder(fhicl::ParameterSet const & p)
:fTracksTag(p.get<std::string>("TracksTag")),
 fHitsTag(p.get<std::string>("HitsTag")),
 fMaxDistance(p.get<double>("window.MaxDistance")),
 fDTime_min(p.get<int>("window.TimeRange.min")*kTDC_PER_USEC),
 fDTime_max(p.get<int>("window.TimeRange.max")*kTDC_PER_USEC)
// Initialize member data here.
{
  produces<novaddt::HitList>();
  mf::LogInfo("MichelEfinder")<<"config:"
    <<"\n TracksTag:"<<fTracksTag
    <<"\n HitsTag:  "<<fHitsTag
    <<"\n window:  "
    <<"\n\t MaxDistance: "<<fMaxDistance
    <<"\n\t DTime limits: ["<<fDTime_min<<","<<fDTime_max<<"]";
}

//-------------------------------------------
void novaddt::MichelEfinder::produce(art::Event & e)
{
  auto Tracks=e.getValidHandle<std::vector<novaddt::HitList>>(fTracksTag);
  auto Hits  =e.getValidHandle<novaddt::HitList>(fHitsTag);

  std::vector<novaddt::DAQHit> trackEnds;
  trackEnds.reserve(Tracks->size());

  mf::LogDebug("MichelEfinder")<<"find tracks ends for Ntrk="<<Tracks->size();
  for(const auto& track: *Tracks){
    auto ends=findtrackends2d(track);
    trackEnds.insert(trackEnds.end(),ends.begin(),ends.end());
  }
  auto result=std::make_unique<novaddt::HitList>();
  findmichelexclusion(trackEnds,*Hits,*result);
 
  mf::LogDebug("MichelEfinder")<<"found "<<result->size()<<" michel hits"
  <<" around "<<trackEnds.size()<<" endpoints";
  e.put(std::move(result));
}

DEFINE_ART_MODULE(novaddt::MichelEfinder)
